{% set class = "md-header" %}
{% if "navigation.tabs.sticky" in features %}
{% set class = class ~ " md-header--shadow md-header--lifted" %}
{% elif "navigation.tabs" not in features %}
{% set class = class ~ " md-header--shadow" %}
{% endif %}
<header class="{{ class }}" data-md-component="header">
    <nav class="md-header__inner md-grid" aria-label="{{ lang.t('header') }}">
        <a href="{{ config.extra.homepage | d(nav.homepage.url, true) | url }}" title="{{ config.site_name | e }}"
            class="md-header__button md-logo" aria-label="{{ config.site_name }}" data-md-component="logo">
            {% include "partials/logo.html" %}
        </a>
        <label class="md-header__button md-icon" for="__drawer">
            {% set icon = config.theme.icon.menu or "material/menu" %}
            {% include ".icons/" ~ icon ~ ".svg" %}
        </label>
        <div class="md-header__title" data-md-component="header-title">
            <div class="md-header__ellipsis">
                <div class="md-header__topic">
                    <span class="md-ellipsis">
                    </span>
                </div>
                <div class="md-header__topic" data-md-component="header-topic">
                    <span class="md-ellipsis">
                    </span>
                </div>
            </div>
        </div>
        {% if config.theme.palette %}
        {% if not config.theme.palette is mapping %}
        {% include "partials/palette.html" %}
        {% endif %}
        {% endif %}
        {% if not config.theme.palette is mapping %}
        {% include "partials/javascripts/palette.html" %}
        {% endif %}
        {% if config.extra.alternate %}
        {% include "partials/alternate.html" %}
        {% endif %}
        {% if "material/search" in config.plugins %}
        <label class="md-header__button md-icon" for="__search">
            {% set icon = config.theme.icon.search or "material/magnify" %}
            {% include ".icons/" ~ icon ~ ".svg" %}
        </label>
        {% include "partials/search.html" %}
        {% endif %}
        {% if config.repo_url %}
        <div class="md-header__source">
            {% include "partials/source.html" %}
        </div>
        {% endif %}
    </nav>
    {% if "navigation.tabs.sticky" in features %}
    {% if "navigation.tabs" in features %}
    {% include "partials/tabs.html" %}
    {% endif %}
    {% endif %}
</header>
